"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
    The GA class provides the evolutionary framework for
    evolving citizens to reach a higher and higher level
    of fitness.
 */
var GA = /** @class */ (function () {
    // Constructor
    //      calcFitness - provide a measure (0 - 1) of how well a citizen performs
    function GA(ga) {
        this.ga = ga;
        // The current generation
        this.generation = 0;
        // The mutation rate for the evolutionary process
        //      A numeric value between 0 and 1
        //      (i.e. what proportion of a citizen's genotype will mutate)
        this.mutationRate = 0.01;
        // The survival rate for the evolutionary process
        //      A numeric value between 0 and 1
        //      (i.e. what proportion of the generation will survive into the next generation)
        this.survivalRate = 0.1;
        // The current population of citizens in the gene pool
        this._population = [];
        this.reset();
    }
    Object.defineProperty(GA.prototype, "population", {
        get: function () {
            return this._population;
        },
        set: function (v) {
            this._population = [].concat(v);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(GA.prototype, "fitness", {
        // The fitness of the current fittest citizen
        //      (based on the fact that the population is sorted by descending fitness)
        get: function () {
            return this._population.length ? this.ga.fitness(this._population[0]) : 0;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(GA.prototype, "fittest", {
        // The current fittest citizen within the population
        //      (based on the fact that the population is sorted by descending fitness)
        get: function () {
            return this._population.length ? this._population[0] : null;
        },
        enumerable: true,
        configurable: true
    });
    // Clear the evolutionary process
    GA.prototype.reset = function () {
        this._population = [];
        this.generation = 0;
    };
    // Sort the population by the fitness of the citizens
    GA.prototype.sort = function () {
        var _this = this;
        this._population.sort(function (a, b) {
            return _this.ga.fitness(b) - _this.ga.fitness(a);
        });
    };
    // Run the evolutionary process for a single generation
    GA.prototype.evolve = function () {
        var nextPopulation = [];
        var populationSize = this._population.length;
        // Create Breeding Pool
        var breeders = [];
        this._population.forEach(function (citizen, idx) {
            // The fitter the citizen, the more often they appear in the breeders array
            for (var i = 0; i < populationSize - idx; i++) {
                breeders.push(citizen);
            }
        });
        // Keep the Best
        var noToKeep = Math.ceil(populationSize * this.survivalRate);
        nextPopulation = nextPopulation.concat(this._population.slice(0, noToKeep));
        // Breed the Rest
        for (var i = 0; i < populationSize - noToKeep; i++) {
            nextPopulation.push(this.ga.mutate(this.ga.mate(breeders[Math.floor(Math.random() * breeders.length)], breeders[Math.floor(Math.random() * breeders.length)]), this.mutationRate));
        }
        // Complete
        this._population = nextPopulation;
        // Sort
        this.sort();
        // Done
        this.generation++;
    };
    return GA;
}());
exports.GA = GA;
